@@ -11,6 +11,7 @@ use Mojo::JSON;
 use Mojo::URL;
 use Mojo::Util 'deprecated';
 use Scalar::Util;
+use Time::Local ();
 
 use constant VALIDATE_HOSTNAME => eval 'require Data::Validate::Domain;1';
 use constant VALIDATE_IP       => eval 'require Data::Validate::IP;1';
@@ -18,7 +19,7 @@ use constant SPECIFICATION_URL => 'http://json-schema.org/draft-04/schema#';
 
 use constant DEBUG => $ENV{JSON_VALIDATOR_DEBUG} || 0;
 
-our $VERSION = '0.95';
+our $VERSION = '0.96';
 our @EXPORT_OK = 'validate_json';
 
 my $BUNDLED_CACHE_DIR = path(path(__FILE__)->dirname, qw(Validator cache));
@@ -41,8 +42,7 @@ has cache_paths => sub {
   return \@paths;
 };
 
-has formats => sub { shift->_build_formats };
-
+has formats  => sub { shift->_build_formats };
 has resolver => sub { \&_resolver };
 
 has ua => sub {
@@ -63,8 +63,9 @@ sub coerce {
 
 sub load_and_validate_schema {
   my ($self, $spec, $args) = @_;
-  my $clone = JSON::Validator->new->schema($spec)->schema->data;
-  my @errors = JSON::Validator->new->schema($args->{schema} || SPECIFICATION_URL)->validate($clone);
+  my $clone = $self->new(%$self)->schema($spec)->schema;
+  my @errors
+    = $self->new(%$self)->schema($args->{schema} || SPECIFICATION_URL)->validate($clone->data);
 
   Carp::confess(join "\n", "Invalid schema:", @errors) if @errors;
   warn "[JSON::Validator] Loaded $spec\n" if DEBUG;
@@ -146,7 +147,7 @@ sub _load_schema {
   # Make sure we create the correct namespace if not already done by Mojo::URL
   $namespace =~ s!#.*$!! if $namespace eq $url;
 
-  return $self->{cached}{$namespace} if $self->{cached}{$namespace};
+  return $doc if $doc = $self->_store(cached => $namespace);
   return eval {
     warn "[JSON::Validator] Loading schema $url namespace=$namespace scheme=$scheme\n"
       if DEBUG;
@@ -229,10 +230,9 @@ sub _register_document {
   $doc = Mojo::JSON::Pointer->new($doc);
   $namespace = Mojo::URL->new($namespace) unless ref $namespace;
   $namespace->fragment(undef);
-
-  $self->{cached}{$namespace} = $doc;
   $doc->data->{id} ||= "$namespace";
-  $self->{cached}{$doc->data->{id}} = $doc;
+  $self->_store(cached => $namespace       => $doc);
+  $self->_store(cached => $doc->data->{id} => $doc);
 
   warn "[JSON::Validator] Register id=$doc->{data}{id} namespace=$namespace\n" if DEBUG;
   return $doc;
@@ -246,50 +246,40 @@ sub _register_document {
 # something up instead using _default_id()
 sub _resolve_schema {
   my ($self, $schema, $namespace) = @_;
-  my (@topics, @refs);
+  my ($resolved, @topics, @refs);
 
   # The if is true if we have already resolved this $namespace. This happens if
   # the same specification has "$ref" with the same value multiple times.
   # There's no need to resolve the same schema twice.
-  return $self->{resolved}{$namespace} if $self->{resolved}{$namespace};
+  return $resolved if $resolved = $self->_store(resolved => $namespace);
 
-  warn "[JSON::Validator] Resolving schema $namespace\n" if DEBUG;
-  $self->{resolved}{$namespace} = Mojo::JSON::Pointer->new({%{$schema->data}});
-  @topics = ($self->{resolved}{$namespace}->data);
+  @topics = ($self->_store(resolved => $namespace, $schema))->data;
 
   # This while loop will traverse the whole specification and and track down
   # each and every "$ref" it finds. @topics is a list of all the data
   # structures that it finds. Note that is start out with just having one item:
   # A hash-ref to the complete specification, but as it goes along, the while()
   # loop will track down more objects and arrays inside the specification and
-  # then loop over those as well, util it has "visited" the whole document.
+  # then loop over those as well, util it has visited the whole document.
   while (@topics) {
     my $topic = shift @topics;
+    next if ref $topic and $self->_seen(topic => $topic);    # Avoid recursion
     if (UNIVERSAL::isa($topic, 'HASH')) {
       for my $k (sort keys %$topic) {
         my $v = $topic->{$k};
-
-        # int($v) returns the memory address of the reference. The "if" below
-        # will not resolve the same "$ref" over again.
-        next if ref $v and $self->{seen}{int($v)}++;
-
-        # Make sure we do not modify the input data structure.
-        # Changing the input makes t/expand.t in swagger2.git fail.
-        $topic->{$k} = [@$v] if UNIVERSAL::isa($v, 'ARRAY');
-        $topic->{$k} = {%$v} if UNIVERSAL::isa($v, 'HASH');
-
         push @topics, $topic->{$k} if ref $topic->{$k};
-        push @refs, $topic if $k eq '$ref' and !ref $v;
+        unshift @refs, $topic if $k eq '$ref' and !ref $v;
       }
     }
     elsif (UNIVERSAL::isa($topic, 'ARRAY')) {
-      push @topics, @$topic;
+      unshift @topics, @$topic;
     }
   }
 
   # Resolve all the "$ref" we found. This will call _resolve_schema() again.
   $self->resolver->($self, $namespace, \@refs);
-  $self->{resolved}{$namespace};
+
+  return $schema;
 }
 
 sub _resolver {
@@ -297,33 +287,49 @@ sub _resolver {
 
   # Seconds step: Resolve $ref
   for my $topic (@$refs) {
-    my $ref = $topic->{'$ref'} or next;    # already resolved?
+    my $ref = delete $topic->{'$ref'} or next;    # already resolved?
     $ref = "#/definitions/$ref" if $ref =~ /^\w+$/;    # TODO: Figure out if this could be removed
+    $ref = Mojo::Util::url_unescape($ref || '');
     $ref = Mojo::URL->new($namespace)->fragment($ref) if $ref =~ s!^\#!!;
     $ref = Mojo::URL->new($ref) unless ref $ref;
 
-    warn "[JSON::Validator] Resolving ref $ref defined in $namespace\n" if DEBUG == 2;
-    my $look_in = $self->{resolved}{$ref->clone->fragment(undef)};
-
+    my $look_in = $self->_store(resolved => $ref);
     if (!$look_in) {
       $look_in = $self->_load_schema($ref, $namespace);
       $look_in = $self->_resolve_schema($look_in, $look_in->data->{id} || $namespace);
-      warn "[JSON::Validator] Will look for $ref in $look_in->{data}{id}\n" if DEBUG == 2;
     }
 
-    $ref = $look_in->get($ref->fragment || '')
-      || die qq[Could not find "$topic->{'$ref'}" ($ref). Typo in schema "$namespace"?];
-    %$topic = %$ref;
-    delete $topic->{id} unless ref $topic->{id};    # TODO: Is this correct?
+    warn "[JSON::Validator] Resolving $ref\n" if DEBUG > 1;
+    warn Data::Dumper::Dumper($look_in->data) if DEBUG and $ref =~ /\b[c]\b/;   # follow the changes
+    my $data = $ref->fragment ? $look_in->get($ref->fragment) : $look_in->data;
+    die qq[Possibly a typo in schema? Could not find "$ref"] unless $data;
+    $topic->{$_} = $data->{$_} for keys %$data;
+    unshift @$refs, $topic if $topic->{'$ref'} and !ref $topic->{'$ref'};
+    delete $topic->{id} if !ref $topic->{id} and $self->isa('JSON::Validator::OpenAPI');
   }
 }
 
+# this code is from Data::Dumper::format_refaddr()
+sub _seen {
+  require Scalar::Util;
+  my $self = shift;
+  my $key = join ':', shift, map { pack 'J', Scalar::Util::refaddr($_) } @_;
+  return $self->{seen}{$key}++;
+}
+
+sub _store {
+  my ($self, $key, $namespace, $schema) = @_;
+  $namespace = Mojo::URL->new($namespace)->fragment(undef)->to_string;
+  return $self->{$key}{$namespace} unless $schema;
+  return $self->{$key}{$namespace} = $schema;
+}
+
 sub _validate {
   my ($self, $data, $path, $schema) = @_;
   my ($type, @errors);
 
   # Avoid recursion
-  return if ref $data and $self->{seen}{"$schema\:$data"}++;
+  return if ref $data and $self->_seen(data => $schema, $data);
 
   # Make sure we validate plain data and not a perl object
   $data = $data->TO_JSON if Scalar::Util::blessed($data) and UNIVERSAL::can($data, 'TO_JSON');
@@ -336,13 +342,13 @@ sub _validate {
   elsif ($type) {
     my $method = sprintf '_validate_type_%s', $type;
     @errors = $self->$method($data, $path, $schema);
-    warn "[JSON::Validator] type @{[$path||'/']} $method [@errors]\n" if DEBUG == 2;
+    warn "[JSON::Validator] type @{[$path||'/']} $method [@errors]\n" if DEBUG > 1;
     return @errors if @errors;
   }
 
   if (my $rules = $schema->{not}) {
     push @errors, $self->_validate($data, $path, $rules);
-    warn "[JSON::Validator] not @{[$path||'/']} == [@errors]\n" if DEBUG == 2;
+    warn "[JSON::Validator] not @{[$path||'/']} == [@errors]\n" if DEBUG > 1;
     return @errors ? () : (E $path, 'Should not match.');
   }
 
@@ -371,7 +377,7 @@ sub _validate_all_of {
     push @expected, $schema_type;
   }
 
-  warn "[JSON::Validator] allOf @{[$path||'/']} == [@errors]\n" if DEBUG == 2;
+  warn "[JSON::Validator] allOf @{[$path||'/']} == [@errors]\n" if DEBUG > 1;
   my $expected = join ' or ', _uniq(@expected);
   return E $path, "allOf failed: Expected $expected, not $type."
     if $expected and @errors + @expected == @$rules;
@@ -387,7 +393,7 @@ sub _validate_any_of {
   for my $rule (@$rules) {
     @e = $self->_validate($data, $path, $rule);
     if (!@e) {
-      warn "[JSON::Validator] anyOf @{[$path||'/']} == success\n" if DEBUG == 2;
+      warn "[JSON::Validator] anyOf @{[$path||'/']} == success\n" if DEBUG > 1;
       return;
     }
     my $schema_type = _guess_schema_type($rule);
@@ -395,7 +401,7 @@ sub _validate_any_of {
     push @expected, $schema_type;
   }
 
-  warn "[JSON::Validator] anyOf @{[$path||'/']} == [@errors]\n" if DEBUG == 2;
+  warn "[JSON::Validator] anyOf @{[$path||'/']} == [@errors]\n" if DEBUG > 1;
   my $expected = join ' or ', _uniq(@expected);
   return E $path, "anyOf failed: Expected $expected, got $type." unless @errors;
   return E $path, sprintf "anyOf failed: %s", _merge_errors(@errors);
@@ -414,11 +420,11 @@ sub _validate_one_of {
   }
 
   if (@errors + @expected + 1 == @$rules) {
-    warn "[JSON::Validator] oneOf @{[$path||'/']} == success\n" if DEBUG == 2;
+    warn "[JSON::Validator] oneOf @{[$path||'/']} == success\n" if DEBUG > 1;
     return;
   }
 
-  if (DEBUG == 2) {
+  if (DEBUG > 1) {
     warn sprintf "[JSON::Validator] oneOf %s == failed=%s/%s / @errors\n", $path || '/',
       @errors + @expected, int @$rules;
   }
@@ -442,7 +448,8 @@ sub _validate_type_enum {
   }
 
   local $" = ', ';
-  return E $path, "Not in enum list: @$enum.";
+  return E $path, sprintf 'Not in enum list: %s.', join ', ',
+    map { ref $_ ? Mojo::JSON::encode_json($_) : $_ } @$enum;
 }
 
 sub _validate_format {
@@ -589,7 +596,7 @@ sub _validate_type_object {
   }
 
   while (my ($k, $r) = each %{$schema->{properties}}) {
-    push @{$rules{$k}}, $r if exists $data->{$k} or $required{$k};
+    push @{$rules{$k}}, $r;
   }
   while (my ($p, $r) = each %{$schema->{patternProperties}}) {
     push @{$rules{$_}}, $r for grep { $_ =~ /$p/ } keys %$data;
@@ -620,10 +627,11 @@ sub _validate_type_object {
 
   for my $k (keys %rules) {
     for my $r (@{$rules{$k}}) {
-      if (!exists $data->{$k} and exists $schema->{default}) {
-        $data->{$k} = $r->{default};
+      if (!exists $data->{$k} and (ref $r eq 'HASH' and exists $r->{default})) {
+
+        #$data->{$k} = $r->{default}; # TODO: This seems to fail when using oneOf and friends
       }
-      else {
+      elsif (exists $data->{$k}) {
         my @e = $self->_validate($data->{$k}, _path($path, $k), $r);
         push @errors, @e;
         push @errors, $self->_validate_type_enum($data->{$k}, _path($path, $k), $r)
@@ -732,8 +740,15 @@ sub _guessed_right {
 }
 
 sub _is_date_time {
-  $_[0] =~ qr/^(\d+)-(\d+)-(\d+)T(\d+):(\d+):(\d+(?:\.\d+)?)(?:Z|([+-])(\d+):(\d+))?$/io;
+  my @time = $_[0]
+    =~ m!^(\d{4})-(\d\d)-(\d\d)[T ](\d\d):(\d\d):(\d\d(?:\.\d+)?)(?:Z|([+-])(\d+):(\d+))?$!io;
+  return 0 unless @time;
+  @time = map { s/^0//; $_ } reverse @time[0 .. 5];
+  $time[4] -= 1;    # month are zero based
+  local $@;
+  return eval { Time::Local::timegm(@time); 1 } || 0;
 }
+
 sub _is_domain { warn "Data::Validate::Domain is not installed"; return; }
 
 sub _is_email {
@@ -813,7 +828,7 @@ JSON::Validator - Validate data against a JSON schema
 
 =head1 VERSION
 
-0.95
+0.96
 
 =head1 SYNOPSIS
 
